/*
 * Canna interface
 *
 * 1996-05-03 Kouichi Hirabayashi
 *
 */

/*
c_open() - open channel to Canna server
c_close() - close channel to Canna server
c_gets(char *buf, int max_length, FILE *ifp) - line input routine
	with Canna using kterm's Status line
c_ready() - returns non-zero value when Canna server is ready to use
*/

#if defined(KANJI) && defined(CANNA)
#include <stdio.h>
#include <canna/jrkanji.h>

#define APPNAME		"mail"
#define MAX_SIZE 1024
#define MAX_WIDTH	(w_col - 2)

static void d_putchar();
static void c_putchar();
static void c_echo();
static void c_guide();
static int c_on();
static int c_off();
static char *c_top();
static char *strsave();

/* term cap */
static char *t_up = "\033[A";
static char *t_ce = "\033[K";
static char *t_mr = "\033[7m";
static char *t_me = "\033[m";
static char *t_ts = "\033[?E\033[?%dT";
static char *t_fs = "\033[?F";
static char *t_ds = "\033[?H";
static int c_erase = 0x08;
static int c_kill = 0x15;

static int _c_ready;
static int c_enable;
static int w_col;
static int w_row;
static jrKanjiStatus ks;
static unsigned char c_buf[MAX_SIZE];
static char *lbufp, lbuf[MAX_SIZE];
static int row, col;
static int c_cnt;
static char c_mode[8];

char *c_gets(src, len, ifp) char *src; FILE *ifp;
{
  int c, c1, c2, i, j, n;
  char *s;

  col = row = 0;
  lbufp = lbuf; *lbufp = '\0';

  if (ifp != stdin || MAX_SIZE < len) {
	fprintf(stderr, "Warning: Program error !\n");
	return fgets(lbuf, sizeof(lbuf), ifp);
  }
  if (src != NULL) {	/* pre-input string */
	strcpy(lbuf, src);
	col = strlen(src);
	lbufp = lbuf + col;
	printf("%s", src);
	col = strlen(src);
  }

  while ((c = getchar()) != EOF) {
	if (c == '\004')
		break;
	if (!c_enable) {
		if (c == 0x0f)	/* ^O */
			c_on();
		else {
			d_putchar(c);
			if (c == '\r' || c == '\n')
				return lbuf;
			continue;
		}
	}
	/* input with Canna */
	if (c == 0x1b) {
		if ((c = getchar()) != 0x5B)
			continue;
		if ((c = getchar()) == 0x41)
			c = CANNA_KEY_Up;	/* up */
		else if (c == 0x42)
			c = CANNA_KEY_Down;	/* down */
		else if (c == 0x43)
			c = CANNA_KEY_Right;	/* right */
		else if (c == 0x44)
			c = CANNA_KEY_Left;	/* left */
		else if (c < 0x31 || 0x36 < c)
			continue;
		else {
			c1 = c;
			if ((c = getchar()) == 0x7e) {
				if (c1 == 0x32)
					c = CANNA_KEY_Insert;
				else if (c1 == 0x35)
					c = CANNA_KEY_Up;
				else if (c1 == 0x36)
					c = CANNA_KEY_Down;
				else
					continue;
			}
			else {
				c2 = c;
				if ((c = getchar()) != 0x7e)
					continue;
				/* Function key */
				if ((c1 == 0x32 && c2 == 0x38) ||
					(c1 == 0x31 && c2 ==  0x31))
					c = CANNA_KEY_Help;	/* F1 */
				else if ((c1 == 0x33 && c2 == 0x33) ||
					(c1 == 0x31 && c2 == 0x32))
					c = CANNA_KEY_Insert;	/* F2 */
				else if ((c1 == 0x33 && c2 == 0x34) ||
					(c1 == 0x31 && c2 == 0x33))
					continue;		/* F3 */
				else if ((c1 == 0x32 && c2 == 0x39) ||
					(c1 == 0x31 && c2 == 0x34))
					c = CANNA_KEY_Xfer;	/* F4 */
				else if (c1 == 0x32 && c2 == 0x31)
					c = CANNA_KEY_Xfer;	/* F10 */
				else
					continue;
			}
		}
	}

	n = jrKanjiString(0, c, c_buf, MAX_SIZE, &ks);
	if (n == 0) {	/* conversion process */
		c_echo();
		if (*c_mode == ' ') {
			c_off();
			continue;
		}
	}
	else if (n > 0) {	/* fixed */
		if (ks.info & KanjiModeInfo) {
			strcpy(c_mode, ks.mode);
			c_guide(ks.mode, NULL);
		}
		c_buf[n] = '\0';
		if (*c_buf == '\r')
			*c_buf = '\n';
		for ( ; c_cnt > 0; c_cnt--)
			d_putchar('\b');
		for (s = c_buf; *s; )
			d_putchar(*s++);
		if (*c_buf == '\n') {
			fflush(stdout);
			return lbuf;
		}
	}
  }
  return NULL;
}

static void d_putchar(c)
{
  static int c0;
  int i, n;
  char *s, *t;

  c &= 0xff;
  if ((c & 0x80) && !c0) {
	c0 = c;
	return;
  }
  if (c == c_erase) {	/* erase */
	t = lbufp - 1;
	for (s = lbuf; s < t; s++)
		if (*s & 0x80)
			s++;
	if (s > t) {	/* KANJI, KANA */
		lbufp -= 2;
		n = ((t[-1] & 0xff) == 0x8e) ? 1 : 2;
	}
	else {
		lbufp--;
		if ((*t & 0xff) < ' ')
			n = 2;
		else
			n = 1;
	}
	if (col > 0) {
		for (i = 0; i < n; i++) {
			putchar('\b');
			col--;
		}
	}
	if (col == 0 && row > 0) {
		putchar('\r');
		printf(t_ce);
		printf(t_up);
		putchar('\r');
		t = c_top(--row, &col);
		for (col, s = t; s < lbufp; s++) {
			if ((*s & 0xff) < ' ') {
				putchar('^');
				putchar(*s + '@');
				col++;
			}
			else
				putchar(*s);
			col++;
		}
	}
	printf(t_ce);
  }
  else if (c == c_kill) {	/* KILL (^O) */
	for (i = 0; i < row; i++) {
		putchar('\r');
		printf(t_ce);
		printf(t_up);
	}
	printf(t_ce);
	row = col = 0;
	*(lbufp = lbuf) = '\0';
  }
  else if (c == '\r' || c == '\n') {
	c = '\n';
	putchar(c);
	*lbufp++ = c;
  }
  else {
	if (col >= MAX_WIDTH) {
		putchar('\r'); putchar('\n');
		row++; col = 0;
	}
	if (c0) {
		putchar(c0);
		if (c0 != 0x8e)
			col++;
		*lbufp++ = c0;
		c0 = 0;
	}
	if (c < ' ') {
		putchar('^');
		putchar(c + '@');
		col++;
	}
	else
		putchar(c);
	col++;
	*lbufp++ = c;
  }
  *lbufp  = '\0';
}

static char *c_top(n, p) int *p;
{
  char *s;
  int c, i, j;

  for (s = lbuf, i = j = 0; i < n && *s; ) {
	if (j >= MAX_WIDTH) {
		i++; j = 0;
		continue;
	}
	if ((c = *s++ & 0xff) & 0x80) {
		if (c != 0x8e)
			j++;
		s++; j++;
	}
	else if (c < ' ')
		j += 2;
	else
		j++;
  }
  *p = j;	/* column */
  return s;
}

static void c_echo()
{
  int i, j;

  char *s, *guide;

  if (ks.info & KanjiModeInfo)
	strcpy(c_mode, ks.mode);
  if (ks.info & KanjiGLineInfo)
	guide = ks.gline.line;
  else
	guide = NULL;
  c_guide(c_mode, guide);
  for ( ; c_cnt > 0; c_cnt--)
	d_putchar('\b');
  if (ks.length > 0) {
	s = ks.echoStr;
	j = ks.revPos;
	d_putchar('|');
	for (i = 0; i < j; i++)
		d_putchar(*s++);
	printf(t_mr);	/* inverse video */
	for (i = 0; i < ks.revLen; i++)
		d_putchar(*s++);
	printf(t_me);
	j = ks.length - ks.revPos - ks.revLen;
	for (i = 0; i < j; i++)
		d_putchar(*s++);
	d_putchar('|');
	for (c_cnt = 2, s = ks.echoStr; *s; c_cnt++, s++)
		if (*s & 0x80)
			*s++;
  }
}

#include <termios.h>

struct termios oldterm, newterm;

c_open()
{
  char **mes, **u, tmps[1024];
  int guidewidth, modewidth, r;
  struct winsize w;
#ifdef TERMCAP
  char *term, tbuf[1024];
  char *s, *t;
#endif

  if (_c_ready)
	return 0;
  if (getenv("DISPLAY") == NULL)
	return -1;
#ifdef TERMCAP
  if ((term = (char *)getenv("TERM")) == NULL)
	return -1;
  if (tgetent(tbuf, term) != 1)
	return -2;
  s = tmps;
  tgetstr("up", &s);
  t_up = strsave(tmps);
  s = tmps;
  tgetstr("ce", &s);
  t_ce = strsave(tmps);
  s = tmps;
  tgetstr("mr", &s);
  t_mr = strsave(tmps);
  s = tmps;
  tgetstr("me", &s);
  t_me = strsave(tmps);
  s = tmps;
  tgetstr("ts", &s);
  t_ts = strsave(tmps);
  /* tricky */
  for (t = t_ts, s = t_ts; *s; )
	if (*s == '%' && s[1] == 'i')
		s += 2;
	else
		*t++ = *s++;
  *t = '\0';
  s = tmps;
  tgetstr("fs", &s);
  t_fs = strsave(tmps);
  s = tmps;
  tgetstr("ds", &s);
  t_ds = strsave(tmps);
#endif

  ioctl(1, TIOCGWINSZ, &w);
  w_row = w.ws_row;
  w_col = w.ws_col;

  if ((r = jrKanjiControl(0, KC_INITIALIZE, (char *)&mes)) == -1) {
	_c_ready = 0;
	return -1;
  }
  if (mes) {
	for (u = mes; *u; u++)
		printf("\007%s\n", *u);
  }
  sprintf(tmps, APPNAME);
  jrKanjiControl(0, KC_SETAPPNAME, tmps);
  modewidth = jrKanjiControl(0, KC_QUERYMAXMODESTR, (char *)0);
  jrKanjiControl(0, KC_QUERYMODE, c_mode);
  c_guide(c_mode, NULL);
  guidewidth = w_col - modewidth;
  jrKanjiControl(0, KC_SETWIDTH, (char *)guidewidth);

  tcgetattr (0, &oldterm);
  tcgetattr (0, &newterm);
#ifdef IGNORE_SIGNAL 
  newterm.c_lflag &= ~(ISIG | ICANON | ECHO | IEXTEN);
#else
  newterm.c_lflag &= ~(ICANON | ECHO | IEXTEN);
#endif
  newterm.c_iflag &= ~ICRNL;
#ifdef NO_ONCR
  newterm.c_oflag &= ~(ONLCR | OPOST);
#endif
  newterm.c_cc[VMIN] = 1;
  newterm.c_cc[VTIME] = 0;
  tcsetattr (0, TCSADRAIN, &newterm);
  c_erase = newterm.c_cc[VERASE];
  c_kill = newterm.c_cc[VKILL];


#ifdef START_KANJI
  ksv.ks = &ks;
  ksv.buffer = c_buf;
  ksv.bytes_buffer = MAX_SIZE;
  ksv.val = CANNA_MODE_HenkanMode;
  ksv.val = CANNA_MODE_AlphaMode;
  jrKanjiControl(0, KC_CHANGEMODE, (char *)&ksv);
  jrKanjiControl(0, KC_QUERYMODE, modeline);
#endif
  c_guide(c_mode, NULL);

  _c_ready = 1;
  return 0;
}

c_close()
{
  if (_c_ready) {
	jrKanjiControl(0, KC_FINALIZE, 0);
	printf(t_ds);	/* disable status line */
	tcsetattr (0, TCSADRAIN, &oldterm);
	_c_ready = 0;
  }
  return 0;
}

c_ready()
{
  return _c_ready;
}

static int c_on()
{
  if (!c_enable)
	c_enable = 1;
  return 0;
}

static int c_off()
{
  jrKanjiStatusWithValue ksv;

  if (c_enable) {
	ksv.ks = &ks;
	ksv.buffer = c_buf;
	ksv.bytes_buffer = MAX_SIZE;
	jrKanjiControl(0, KC_KAKUTEI, (char *)&ksv);
	c_enable = 0;
  }
  return 0;
}

static void c_guide(mode, guide) char *mode, *guide;
{
  char *u, tmps[32];
  int i, j;

  sprintf(tmps, t_ts, 1);	/* goto Status Line */
  printf("%s", tmps);
  printf("%s", t_mr);
  if (strcmp("      ", mode))
	fputs(mode, stdout);
  if (guide != NULL) {
	u = ks.gline.line;
	j = ks.gline.revPos;
	for (i = 0; i < j; i++)
		putchar(*u++);
	printf("%s", t_me);
	for (i = 0; i < ks.gline.revLen; i++)
		putchar(*u++);
	printf("%s", t_mr);
	j = ks.gline.length - ks.gline.revPos - ks.gline.revLen;
	for (i = 0; i < j; i++)
		putchar(*u++);
#if 0
	printf(t_ce);	/* clear to EOL */
#endif
  }
  printf(t_fs);
}
static char *strsave(s) char *s;
{
  char *p;

  if ((p = (char *)malloc(strlen(s) + 1)) != NULL)
	strcpy(p, s);
  return p;
}
#endif	/* KANJI, CANNA */
